setInfos(6, "8-Puzzle", "The 8-, 15- Sliding Puzzle Game", "wodahs", "V1.0", 0xFFAAFF, "en", ["game"])

show(null)
setImage(null)

def imageList = ["wodahs/8Puzzle/1.jpg", "wodahs/8Puzzle/2.jpg", "wodahs/8Puzzle/3.jpg"]

def debugText = ""
def debuginfo = false
def showShuffle = false
def playImage
def playSize
def playArr
def shuffleCount
def slideCount = 0
def startTime
def endTime
def usedTime
def text = ""
def local


switch(System.getProperty("user.language")) {
  case "de":
    local = [
      left: "Links",
      right: "Rechts",
      up: "Oben",
      down: "Unten",
      quit: "Ende",
      yes: "Ja",
      no: "Nein",
      bye: "Wiedersehen!",
      again: "Noch ein Spiel?",
      shuffle: "Mischen",
      showShuffle: "Mischen mit zuschauen",
      chooseImage: "Bitte ein Bild auswählen",
      selSize: "Bitte eine Spielfeldgröße wählen",
      gratz: "Glückwunsch!",
      slides: "Züge",
      time: "Zeit",
      min: "Minuten",
      s3: "3 (Anfänger)",
      s4: "4 (Normal)",
      s5: "5 (Fortgeschritten)",
      s6: "6 (Experte)",
      s7: "7 (Profi)",
      imageQuestion: "Mit diesem Bild spielen?",
      chooseDir: "Richtung wählen",
      score: "Deine Punkte",
      highscore: "Dein Highscore",
      newHighscore: "Neuer Highscore!"
    ]
    break
  default:
    local = [
      left: "Left",
      right: "Right",
      up: "Up",
      down: "Down",
      quit: "Quit",
      yes: "Yes",
      no: "No",
      bye: "Bye!",
      again: "Play again?",
      shuffle: "shuffle",
      showShuffle: "spectate at shuffle",
      chooseImage: "Choose Image",
      selSize: "Select Size",
      gratz: "Congratulations!",
      slides: "Slides",
      time: "Time",
      min: "minutes",
      s3: "3 (beginner)",
      s4: "4 (standard)",
      s5: "5 (advanced)",
      s6: "6 (expert)",
      s7: "7 (professional)",
      imageQuestion: "Do you want to play with it?",
      chooseDir: "Choose direction",
      score: "Your Score",
      highscore: "Your Highscore",
      newHighscore: "New Highscore!"
    ]
    break
}

def init = { key, value ->
	def tmp = loadString(key)
	if (tmp == null)
		return value
	else
		return tmp
}




/*///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	8-Puzzle functions
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////*/

def getIndex = { x, y, size ->
  return size * y + x
}


def getX = { index, size ->
  return index % size
}


def getY = { index, size ->
  return (int)(index / size)
}


def getXY = { index, size ->
  return [getX(index, size), getY(index, size)]
}


def showAll = { arr, size ->
  for (def i = 0; i < arr.size(); i++) {
    arr[i][1] = true
  }
}

def generateArr = { size ->
    def arr = []
    def showState = true
    for (def i = 0; i < size * size; i++)
    {
      if (i == size * size -1) showState = false
      arr.add([i, showState])
    }
    return arr
}

def showTiledImageV2 = { picture, arr, spacing, size ->
  def image = java.awt.Toolkit.getDefaultToolkit().getImage("images/" + picture)
  def width = image.getWidth()
  def height = image.getHeight()
  def tileWidth = (int)(width / size)
  def tileHeight = (int)(height / size)
  //aufgrund rundung entstehen am rechten und unteren rand schwarze ränder

  def frameBuffer = new java.awt.image.BufferedImage(width + spacing , height + spacing, java.awt.image.BufferedImage.TYPE_4BYTE_ABGR)
  java.awt.Graphics2D graphics = frameBuffer.createGraphics()

  graphics.setBackground(java.awt.Color.BLACK)
  graphics.clearRect(0, 0, frameBuffer.getWidth(), frameBuffer.getHeight())

  for (def i = 0; i < arr.size(); i++)
  {
    if (arr[i][1])
    {
      def (x, y) = getXY(arr[i][0], size)

      graphics.drawImage(image,

        tileWidth * getX(i, size) + spacing,
        tileHeight * getY(i, size) + spacing,
        tileWidth * (getX(i, size) +1),
        tileHeight * (getY(i, size) +1),

        tileWidth * x  ,
        tileHeight * y ,
        tileWidth * (x +1),
        tileHeight * (y +1),

        null)
    }
  }

  java.io.ByteArrayOutputStream baos = new java.io.ByteArrayOutputStream()
  javax.imageio.ImageIO.write(frameBuffer, "png", baos)
  setImage(baos.toByteArray(), 0)
}


def getFreeTileIndex = { arr ->
  def freeTileIndex = -1
  for (def i = 0; i < arr.size(); i++)
  {
    if (!arr[i][1]) {
        freeTileIndex = i
        if (debuginfo) debugText += "<br>freeIndex: " +i + ","
        break
      }
  }
  return freeTileIndex
}


def getpossibleNeighbours = { freeTileIndex, size ->
  def possibleNeighbours = []
  def (x, y) = getXY(freeTileIndex, size)

  //Oben
  if (y > 0) possibleNeighbours.add([x,y-1])
  //Unten
  if (y < size -1) possibleNeighbours.add([x,y+1])
  //Links
  if (x > 0) possibleNeighbours.add([x-1,y])
  //Rechts
  if (x < size -1) possibleNeighbours.add([x+1,y])

  return possibleNeighbours
}


def shuffleArray = { count, arr, size ->
  def lastSwapIndex = -1
  for (count; count > 0; count--)
  {
    if (debuginfo) debugText += "Count: " + count

    // freies Feld finden und merken
    def freeTileIndex = getFreeTileIndex(arr)


    // mögliche nachtbaren finden und merken
    def possibleNeighbours = getpossibleNeighbours(freeTileIndex, size)
    if (debuginfo) debugText += "<br>pn: " + possibleNeighbours


    // einen nachtbaren auswählen und feld tauschen
    def swapPartner = possibleNeighbours[getRandom(possibleNeighbours.size())]
    def swapPartnerIndex = getIndex(swapPartner[0], swapPartner[1], size)
    while (swapPartnerIndex == lastSwapIndex)
    {
      swapPartner = possibleNeighbours[getRandom(possibleNeighbours.size())]
      swapPartnerIndex = getIndex(swapPartner[0], swapPartner[1], size)
    }
    if (debuginfo) debugText += "<br>lsp: " + lastSwapIndex + "  sp: " +swapPartnerIndex

    def swapTmp = arr[freeTileIndex]
    arr[freeTileIndex] = arr[swapPartnerIndex]
    arr[swapPartnerIndex] = swapTmp


    //letzen tauschpartner merken
    lastSwapIndex = freeTileIndex


    if (debuginfo) debugText += "<br>" + arr + "<br>"

    //debug Bild anzeigen wähend mischen
    if (showShuffle) showTiledImageV2(playImage, arr, 1, size)
    if (debuginfo) {
      show(debugText)
      debugText = ""
    }
  }

  return arr
}


def contains = { arr, x, y ->
  def ret = false
  for (def i = 0; i < arr.size(); i++) {
    if (arr[i][0] == x && arr[i][1] == y) {
       ret = true
       break
    }
  }
  return ret
}


def userSwap = { direction, arr, size ->
  //0=links, 1=rechts, 2=oben, 3=unten
  debugText = ""

  //usereingabe prüfen ob erlaubt
  def allowed = false
  def freeTileIndex = getFreeTileIndex(arr)
  def swapPartnerIndex = -1
  def possibleNeighbours = getpossibleNeighbours(freeTileIndex, size)
  if (debuginfo) debugText += "<br>pn: " +possibleNeighbours

  switch(direction) {
    case 0:
      allowed = contains(possibleNeighbours, getX(freeTileIndex, size) -1, getY(freeTileIndex, size))
      swapPartnerIndex = getIndex(getX(freeTileIndex, size) -1, getY(freeTileIndex, size), size)
    break
    case 1:
      allowed = contains(possibleNeighbours, getX(freeTileIndex, size) +1, getY(freeTileIndex, size))
      swapPartnerIndex = getIndex(getX(freeTileIndex, size) +1, getY(freeTileIndex, size), size)
    break
    case 2:
      allowed = contains(possibleNeighbours, getX(freeTileIndex, size), getY(freeTileIndex, size) -1)
      swapPartnerIndex = getIndex(getX(freeTileIndex, size), getY(freeTileIndex, size) -1, size)
    break
    case 3:
      allowed = contains(possibleNeighbours, getX(freeTileIndex, size), getY(freeTileIndex, size) +1)
      swapPartnerIndex = getIndex(getX(freeTileIndex, size), getY(freeTileIndex, size) +1, size)
    break
  }
  if (debuginfo) debugText += "<br>allowed: " + allowed

  if (!allowed) return


  //swap durchführend
  def swapTmp = arr[freeTileIndex]
  arr[freeTileIndex] = arr[swapPartnerIndex]
  arr[swapPartnerIndex] = swapTmp


  //prüfen ob puzzel fertig
  def finish = true
  for (def i = 0; i < arr.size(); i++) {
    if (arr[i][0] != i) finish = false
  }
  if (debuginfo) debugText += "<br>" + arr
  return finish
}


/*///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	Menue functions
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////*/

def chooseImage = { arr ->
  def menueButton = []
  for (def i = 0; i < arr.size(); i++)
  {
    menueButton.add(arr[i])
  }
  def wahl = getSelectedValue(local.chooseImage,	menueButton)
  setImage(arr[wahl])
  while (getSelectedValue(local.imageQuestion, [local.yes, local.no]) == 1) {
    wahl = getSelectedValue(local.chooseImage,	menueButton)
    setImage(arr[wahl])
  }
  show(null)
  return arr[wahl]
}


def chooseSize = {
  def size = getSelectedValue(local.selSize,	[local.s3, local.s4, local.s5, local.s6, local.s7]) + 3
  show(null)
  return size
}


def reset = {
  setImage(null)
  slideCount = 0
}


def checkHighscore = { size, score ->
  def ret = 0
  def saveKey = "wodahs.8Puzzle.highscore.s" + size
  def highscore = init(saveKey, 0).toInteger()
  if (score > highscore) {
    save(saveKey, score)
    ret = highscore
  }
  return ret
}

/*///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	8-Puzzle Main
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////*/

def playing = true
while (playing)
{
  playImage = chooseImage(imageList)
  playSize = chooseSize()
  shuffleCount = playSize * playSize * 3
  playArr = generateArr(playSize)

  //showTiledImageV2(playImage, playArr, 1, playSize) //manchmal wird erstes anzeigen nur schwarz angezeigt, nochmal aufrufen behebt problem
  showTiledImageV2(playImage, playArr, 1, playSize)
  if (getSelectedValue("", [local.shuffle, local.showShuffle]) == 1) showShuffle = true
  shuffleArray(shuffleCount, playArr, playSize)
  showTiledImageV2(playImage, playArr, 1, playSize)

  startTime = getTime()
  def roundRunning = true
  def menutext = local.chooseDir
  while (roundRunning)
  {
    if (debuginfo) menutext = debugText
    def wahl = getSelectedValue(menutext,	[local.left, local.right, local.up, local.down, local.quit])
    if (wahl == 4) {
      // QUIT
      roundRunning = false
      playing = false
    }
    else {
      // GAME
      def finish = userSwap(wahl, playArr, playSize)
      showTiledImageV2(playImage, playArr, 1, playSize)
      slideCount++
      if (finish) {
        roundRunning = false
        endTime = getTime()
        usedTime = (int)((endTime - startTime) / 60)
        showAll(playArr, playSize)
        showTiledImageV2(playImage, playArr, 0, playSize)
        text = "${local.gratz}<br><br>${local.slides}: " + slideCount +"<br>${local.time}: " +usedTime + " ${local.min}<br><br>"
        def score = Math.round(1000 - (slideCount * 2) + (usedTime * 10))
        def highscore = checkHighscore(playSize, score)
        if (highscore > 0) {
          text += local.newHighscore + "<br>${local.score}: " + score
        }
        else {
          text += "${local.score}: " + score + "    (${local.highscore}: " + highscore + ")"
        }
        text += "<br><br><br>"
        switch(getSelectedValue(text + local.again,	[local.yes, local.quit])) {
          case 0:
            reset()
            break
          default:
            playing = false
            break
        }
      }
    }
  }
}

show(text + local.bye)
return null

// This script, from http://ss.deviatenow.com, is protected by this licence :
// CC by-NC, see http://creativecommons.org/licenses/by-nc/3.0/
// (Author may change thoses lines)

/*
 * Resources
 * scripts/wodahs/8Puzzle.groovy
 * images/wodahs/8Puzzle/1.jpg
 * images/wodahs/8Puzzle/2.jpg
 * images/wodahs/8Puzzle/3.jpg
*/
